/* 
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.pdfextractor.structure;

import java.util.ArrayList;
import java.util.Objects;

/**
 * This represents a syntatic block of Text, which can be a column on a page, a header or something
 * similar. Formally it is defined as a sequence of paragraphs, where the number of paragraphs might
 * very well be only one.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class TextBlock extends LineSegment {

	/**
	 * This is the actual content of the TextBlock.
	 */
	public final ArrayList<Paragraph> content = new ArrayList<>();
	private double relativeFontSize;

	public TextBlock() {
	}

	public double getRelativeFontSize() {
		return relativeFontSize;
	}

	public void setRelativeFontSize(double relativeFontSize) {
		this.relativeFontSize = relativeFontSize;
	}

	@Override
	public int hashCode() {
		int hash = 5;
		hash = 67 * hash + Objects.hashCode(this.content);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final TextBlock other = (TextBlock) obj;
		if (!Objects.equals(this.content, other.content)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		final StringBuilder output = new StringBuilder();
		for (int i = 0; i < content.size(); i++) {
			output.append('\t');
			output.append(content.get(i).toString());
			if (i < content.size() - 1) {
				output.append("\n\n");
			}
		}
		return output.toString();
	}

	/**
	 * Does the same as toString but also inserts the beginning and end index of each objects
	 * respective text representation.
	 */
	public String indexedToString(int currentIdx) {
		setBegin(currentIdx);
		final StringBuilder output = new StringBuilder();
		for (int i = 0; i < content.size(); i++) {
			output.append('\t');
			currentIdx++;
			final String parStr = content.get(i).indexedToString(currentIdx);
			output.append(parStr);
			currentIdx += parStr.length();
			if (i < content.size() - 1) {
				output.append("\n\n");
				currentIdx += 2;
			}
		}
		setEnd(currentIdx);
		return output.toString();
	}

	public String toXML() {
		final StringBuilder output = new StringBuilder();
		output.append("\t\t<TextBlock RELATIVE_FONT_SIZE=\"");
		output.append(Double.toString(relativeFontSize));
		output.append("\">\n");
		for (int i = 0; i < content.size(); i++) {
			output.append(content.get(i).toXML());
			if (i < content.size() - 1) {
				output.append("\n");
			}
		}
		output.append("\n\t\t</TextBlock>");
		return output.toString();
	}
}
